<?php
/*
  set_include_path(
    dirname(__DIR__) . DIRECTORY_SEPARATOR . 'vendor/pear/db' . PATH_SEPARATOR .
    dirname(__DIR__) . PATH_SEPARATOR .
    get_include_path()
  );
*/
$civicrm_install_path = '/usr/share/civicrm';
set_include_path(
    $civicrm_install_path . PATH_SEPARATOR
    .$civicrm_install_path . DIRECTORY_SEPARATOR . 'vendor' . PATH_SEPARATOR
    .$civicrm_install_path . DIRECTORY_SEPARATOR . 'vendor/pear/db' . PATH_SEPARATOR
    .$civicrm_install_path . DIRECTORY_SEPARATOR . 'vendor/pear/log' . PATH_SEPARATOR
    .$civicrm_install_path . DIRECTORY_SEPARATOR . 'packages' . PATH_SEPARATOR
    .get_include_path()
);

  require_once 'Log/observer.php';					// php-log

  require_once dirname(__DIR__) . '/Civi.php';
  require_once dirname(__DIR__) . '/Civi/Core/HookInterface.php';
  require_once dirname(__DIR__) . '/Civi/Test/HookInterface.php';

//require_once 'Composer/Autoload/ClassLoader.php';
  require_once 'composer/ClassLoader.php';


  require_once 'league/csv/autoload.php';
  require_once 'xkerman/restricted-unserialize/src/function.php';

  require_once 'Psr/Log/autoload.php';					// php-psr-log
  require_once 'Psr/SimpleCache/autoload.php';				// php-psr-simple-cache
  require_once 'HTMLPurifier.auto.php';					// php-htmlpurifier
  require_once 'phpseclib.autoloader.php';				// php-seclib

  require_once 'guzzlehttp/guzzle/src/functions_include.php';
  require_once 'guzzlehttp/promises/src/functions_include.php';
  require_once 'guzzlehttp/psr7/src/functions_include.php';
//require_once 'GuzzleHttp/Promise/autoload.php';			// php-guzzlehttp-promises
//require_once 'GuzzleHttp/Promise/functions_include.php';		// php-guzzlehttp-promises
//require_once 'GuzzleHttp/Psr7/functions_include.php';			// php-guzzlehttp-psr7

  require_once 'Symfony/Component/Config/autoload.php';			// php-symfony-config
//Incompatible with Symfony Component(s) 4+:
//require_once 'Symfony/Component/DependencyInjection/autoload.php';	// php-symfony-dependency-injection
//require_once 'Symfony/Component/EventDispatcher/autoload.php';	// php-symfony-event-dispatcher
//require_once 'Symfony/Component/Filesystem/autoload.php';		// php-symfony-filesystem
//require_once 'Symfony/Component/Finder/autoload.php';			// php-symfony-finder
//require_once 'Symfony/Component/Process/autoload.php';		// php-symfony-process

## Crucial for "Fetch Bounces" job.
require_once 'ezc/Base/autoload.php';					// php-zeta-base

/*
  require_once 'packages/smarty3/vendor/smarty/smarty/libs/Autoloader.php';
  require_once 'Auth/SASL.php';						// php-auth-sasl
//require_once 'Cache/IntegrationTests/autoload.php';			// php-cache-integration-tests
  require_once 'Cache/TagInterop/autoload.php';				// php-cache-tag-interop
  require_once 'Console/Getopt.php';					// php-pear
//require_once 'dompdf/dompdf.php';					// php-dompdf
  require_once 'FontLib/Autoloader.php';				// php-font-lib
  require_once 'HTMLPurifier.auto.php';					// php-htmlpurifier
//require_once 'Mail/mime.php';						// php-mail-mime
  require_once 'Mail.php';						// php-mail
  require_once 'Net/SMTP.php';						// php-net-smtp
  require_once 'Net/Socket.php';					// php-net-socket
//require_once 'pear/db/DB.php';					// php-db :: bundled
  require_once 'PEAR/Exception.php';					// php-pear
  require_once 'phpseclib.autoloader.php';				// php-seclib
  require_once 'Psr/Cache/autoload.php';				// php-psr-cache
//require_once 'Psr/Http/Client/autoload.php';				// php-psr-http-client
//require_once 'Psr/Http/Message/autoload.php';				// php-psr-http-message
  require_once 'tcpdf/autoload.php';					// php-tcpdf
  require_once 'Zend/Stdlib/autoload.php';				// php-zend-stdlib
*/

// @codingStandardsIgnoreFile
// @codeCoverageIgnoreStart
// this is an autogenerated file - do not edit
spl_autoload_register(
    function($class) {
        static $classes = null;
        if ($classes === null) {
            $classes = array(
                ___CLASSLIST___
            );
        }
        if (isset($classes[$class])) {
            require __DIR__ . $classes[$class];
        }
    },
    true,
    true
);
// @codeCoverageIgnoreEnd
